@@ -3,10 +3,11 @@ from typing import Iterable, List
 import torch
 from torch.optim import Optimizer
 
+from pytorch_optimizer.base_optimizer import BaseOptimizer
 from pytorch_optimizer.types import BETAS, CLOSURE, DEFAULTS, LOSS, PARAMETERS
 
 
-class AdaHessian(Optimizer):
+class AdaHessian(Optimizer, BaseOptimizer):
     """
     Reference : https://github.com/davda54/ada-hessian
     Example :
@@ -59,7 +60,7 @@ class AdaHessian(Optimizer):
         self.eps = eps
         self.seed = seed
 
-        self.check_valid_parameters()
+        self.validate_parameters()
 
         # use a separate generator that deterministically generates
         # the same `z`s across all GPUs in case of distributed training
@@ -79,19 +80,12 @@ class AdaHessian(Optimizer):
             p.hess = 0.0
             self.state[p]['hessian_step'] = 0
 
-    def check_valid_parameters(self):
-        if self.lr < 0.0:
-            raise ValueError(f'Invalid learning rate : {self.lr}')
-        if self.weight_decay < 0.0:
-            raise ValueError(f'Invalid weight_decay : {self.weight_decay}')
-        if not 0.0 <= self.betas[0] < 1.0:
-            raise ValueError(f'Invalid beta_0 : {self.betas[0]}')
-        if not 0.0 <= self.betas[1] < 1.0:
-            raise ValueError(f'Invalid beta_1 : {self.betas[1]}')
-        if not 0.0 <= self.hessian_power <= 1.0:
-            raise ValueError(f'Invalid hessian_power : {self.hessian_power}')
-        if self.eps < 0.0:
-            raise ValueError(f'Invalid eps : {self.eps}')
+    def validate_parameters(self):
+        self.validate_learning_rate(self.lr)
+        self.validate_betas(self.betas)
+        self.validate_weight_decay(self.weight_decay)
+        self.validate_hessian_power(self.hessian_power)
+        self.validate_epsilon(self.eps)
 
     def get_params(self) -> Iterable:
         """Gets all parameters in all param_groups with gradients"""
@@ -104,7 +98,9 @@ class AdaHessian(Optimizer):
                 p.hess.zero_()
 
     def set_hessian(self):
-        """Computes the Hutchinson approximation of the hessian trace and accumulates it for each trainable parameter"""
+        """Computes the Hutchinson approximation of the hessian trace
+        and accumulates it for each trainable parameter
+        """
         params = []
         for p in self.get_params():
             if p.grad is None:
@@ -142,10 +138,22 @@ class AdaHessian(Optimizer):
                 # approximate the expected values of z * (H@z)
                 p.hess += h_z * z / self.num_samples
 
+    @torch.no_grad()
+    def reset(self):
+        for group in self.param_groups:
+            for p in group['params']:
+                state = self.state[p]
+
+                state['step'] = 0
+                state['exp_avg'] = torch.zeros_like(p)
+                state['exp_hessian_diag_sq'] = torch.zeros_like(p)
+
+    @torch.no_grad()
     def step(self, closure: CLOSURE = None) -> LOSS:
         loss: LOSS = None
         if closure is not None:
-            loss = closure()
+            with torch.enable_grad():
+                loss = closure()
 
         self.zero_hessian()
         self.set_hessian()
@@ -164,8 +172,8 @@ class AdaHessian(Optimizer):
                 state = self.state[p]
                 if len(state) == 1:
                     state['step'] = 0
-                    state['exp_avg'] = torch.zeros_like(p.data)
-                    state['exp_hessian_diag_sq'] = torch.zeros_like(p.data)
+                    state['exp_avg'] = torch.zeros_like(p)
+                    state['exp_hessian_diag_sq'] = torch.zeros_like(p)
 
                 exp_avg, exp_hessian_diag_sq = state['exp_avg'], state['exp_hessian_diag_sq']
 
@@ -173,20 +181,18 @@ class AdaHessian(Optimizer):
                 beta1, beta2 = group['betas']
 
                 # Decay the first and second moment running average coefficient
-                exp_avg.mul_(beta1).add_(p.grad, alpha=1 - beta1)
-                exp_hessian_diag_sq.mul_(beta2).addcmul_(p.hess, p.hess, value=1 - beta2)
+                exp_avg.mul_(beta1).add_(p.grad, alpha=1.0 - beta1)
+                exp_hessian_diag_sq.mul_(beta2).addcmul_(p.hess, p.hess, value=1.0 - beta2)
 
-                bias_correction1 = 1 - beta1 ** state['step']
-                bias_correction2 = 1 - beta2 ** state['step']
+                bias_correction1 = 1.0 - beta1 ** state['step']
+                bias_correction2 = 1.0 - beta2 ** state['step']
 
-                hessian_power = group['hessian_power']
-                denom = (exp_hessian_diag_sq / bias_correction2).pow_(hessian_power / 2).add_(group['eps'])
+                de_nom = (exp_hessian_diag_sq / bias_correction2).pow_(group['hessian_power'] / 2.0).add_(group['eps'])
 
-                if group['adamd_debias_term']:
-                    step_size = group['lr']
-                else:
-                    step_size = group['lr'] / bias_correction1
+                step_size = group['lr']
+                if not group['adamd_debias_term']:
+                    step_size /= bias_correction1
 
-                p.addcdiv_(exp_avg, denom, value=-step_size)
+                p.addcdiv_(exp_avg, de_nom, value=-step_size)
 
         return loss
