@@ -3,8 +3,8 @@ import torch
 
 def centralize_gradient(x: torch.Tensor, gc_conv_only: bool = False) -> torch.Tensor:
     """Gradient Centralization (GC)
-    :param x: torch.Tensor. gradient.
-    :param gc_conv_only: bool. 'False' for both conv & fc layers.
+    :param x: torch.Tensor. gradient
+    :param gc_conv_only: bool. 'False' for both conv & fc layers
     :return: torch.Tensor. GC-ed gradient
     """
     size: int = x.dim()
