@@ -6,6 +6,7 @@ import torch.nn.functional as F
 from torch.optim import Optimizer
 
 from pytorch_optimizer.agc import agc
+from pytorch_optimizer.base_optimizer import BaseOptimizer
 from pytorch_optimizer.gc import centralize_gradient
 from pytorch_optimizer.types import BETAS, CLOSURE, DEFAULTS, LOSS, PARAMETERS, STATE
 from pytorch_optimizer.utils import normalize_gradient, unit_norm
@@ -20,7 +21,7 @@ __AUTHORS__ = [
 ]
 
 
-class Ranger21(Optimizer):
+class Ranger21(Optimizer, BaseOptimizer):
     """
     Reference 1 : https://github.com/lessw2020/Ranger21
     Reference 2 : https://github.com/nestordemeure/flaxOptimizers/blob/main/flaxOptimizers/ranger21.py
@@ -59,7 +60,7 @@ class Ranger21(Optimizer):
         norm_loss_factor: float = 1e-4,
         eps: float = 1e-8,
     ):
-        """Ranger21
+        """Ranger21 optimizer
         :param params: PARAMETERS. iterable of parameters to optimize or dicts defining parameter groups
         :param lr: float. learning rate
         :param beta0: float. Manages the amplitude of the noise introduced by positive negative momentum
@@ -92,7 +93,7 @@ class Ranger21(Optimizer):
         self.norm_loss_factor = norm_loss_factor
         self.eps = eps
 
-        self.check_valid_parameters()
+        self.validate_parameters()
 
         # lookahead
         self.lookahead_step: int = 0
@@ -102,8 +103,6 @@ class Ranger21(Optimizer):
         self.current_lr = lr
         self.min_lr = warm_down_min_lr
 
-        self.param_size: int = 0
-
         defaults: DEFAULTS = dict(
             lr=lr,
             betas=betas,
@@ -126,28 +125,34 @@ class Ranger21(Optimizer):
         self.start_warm_down: int = num_iterations - self.num_warm_down_iterations
         self.warm_down_lr_delta: float = self.starting_lr - self.min_lr
 
-    def check_valid_parameters(self):
-        if self.lr < 0.0:
-            raise ValueError(f'Invalid learning rate : {self.lr}')
-        if not 0.0 <= self.betas[0] < 1.0:
-            raise ValueError(f'Invalid beta_0 : {self.betas[0]}')
-        if not 0.0 <= self.betas[1] < 1.0:
-            raise ValueError(f'Invalid beta_1 : {self.betas[1]}')
-        if self.weight_decay < 0.0:
-            raise ValueError(f'Invalid weight_decay : {self.weight_decay}')
-        if self.eps < 0.0:
-            raise ValueError(f'Invalid eps : {self.eps}')
+    def validate_parameters(self):
+        self.validate_learning_rate(self.lr)
+        self.validate_betas(self.betas)
+        self.validate_beta0(self.beta0)
+        self.validate_weight_decay(self.weight_decay)
+        self.validate_epsilon(self.eps)
 
-    def __setstate__(self, state: STATE):
-        super().__setstate__(state)
+    @torch.no_grad()
+    def reset(self):
+        for group in self.param_groups:
+            for p in group['params']:
+                state = self.state[p]
+
+                state['step'] = 0
+                state['grad_ma'] = torch.zeros_like(p)
+                state['variance_ma'] = torch.zeros_like(p)
+                state['lookahead_params'] = torch.empty_like(p)
+                state['lookahead_params'].copy_(p)
+                state['neg_grad_ma'] = torch.zeros_like(p)
+                state['max_variance_ma'] = torch.zeros_like(p)
 
     @staticmethod
     def build_warm_up_iterations(total_iterations: int, beta2: float, warm_up_pct: float = 0.22) -> int:
-        beta_warm_up_iterations: int = math.ceil(2.0 / (1.0 - beta2))  # default un-tuned linear warmup
-        beta_pct: float = beta_warm_up_iterations / total_iterations
+        warm_up_iterations: int = math.ceil(2.0 / (1.0 - beta2))  # default un-tuned linear warmup
+        beta_pct: float = warm_up_iterations / total_iterations
         if beta_pct > 0.45:
             return int(warm_up_pct * total_iterations)
-        return beta_warm_up_iterations
+        return warm_up_iterations
 
     @staticmethod
     def build_warm_down_iterations(total_iterations: int, warm_down_pct: float = 0.72) -> int:
@@ -182,13 +187,6 @@ class Ranger21(Optimizer):
 
         return new_lr
 
-    @staticmethod
-    def get_state_values(group, state: STATE):
-        beta1, beta2 = group['betas']
-        mean_avg = state['mean_avg']
-        variance_avg = state['variance_avg']
-        return beta1, beta2, mean_avg, variance_avg
-
     @torch.no_grad()
     def step(self, closure: CLOSURE = None) -> LOSS:
         loss: LOSS = None
@@ -205,7 +203,7 @@ class Ranger21(Optimizer):
                 if p.grad is None:
                     continue
 
-                grad = p.grad.data
+                grad = p.grad
                 if grad.is_sparse:
                     raise RuntimeError('Ranger21 does not support sparse gradients')
 
@@ -214,17 +212,13 @@ class Ranger21(Optimizer):
                 # Apply Adaptive Gradient Clipping (AGC)
                 agc(p, agc_eps=self.agc_eps, agc_clip_val=self.agc_clipping_value)
 
-                grad = p.grad
-                if grad.is_sparse:
-                    raise RuntimeError('sparse matrix not supported atm')
-
                 state = self.state[p]
                 if len(state) == 0:
                     state['step'] = 0
                     state['grad_ma'] = torch.zeros_like(p)
                     state['variance_ma'] = torch.zeros_like(p)
-                    state['lookahead_params'] = torch.zeros_like(p.data)
-                    state['lookahead_params'].copy_(p.data)
+                    state['lookahead_params'] = torch.empty_like(p)
+                    state['lookahead_params'].copy_(p)
                     state['neg_grad_ma'] = torch.zeros_like(p)
                     state['max_variance_ma'] = torch.zeros_like(p)
 
@@ -237,17 +231,14 @@ class Ranger21(Optimizer):
 
                 beta1, beta2 = group['betas']
 
-                bias_correction2 = 1 - beta2 ** state['step']
+                bias_correction2 = 1.0 - beta2 ** state['step']
 
                 # second moment estimation
                 # using positive-negative momentum and bias correction
                 variance_ma = state['variance_ma']
-                variance_ma.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
+                variance_ma.mul_(beta2).addcmul_(grad, grad, value=1.0 - beta2)
                 variance_ma_sum += (variance_ma / bias_correction2).sum()
 
-        if not self.param_size:
-            self.param_size = param_size
-
         # stable weight decay
         variance_normalized = math.sqrt(variance_ma_sum / param_size)
         if math.isnan(variance_normalized):
@@ -255,7 +246,8 @@ class Ranger21(Optimizer):
 
         # Phase 2 - Apply weight decay and step
         for group in self.param_groups:
-            if group['params'][0].grad is None:
+            p = group['params'][0]
+            if p.grad is None:
                 continue
 
             lr = group['lr']
@@ -272,11 +264,11 @@ class Ranger21(Optimizer):
             # stable decay
             decay = group['weight_decay']
             if decay:
-                p.data.mul_(1.0 - decay * lr / variance_normalized)
+                p.mul_(1.0 - decay * lr / variance_normalized)
 
             # norm loss
-            u_norm = unit_norm(p.data)
-            correction = 2.0 * self.norm_loss_factor * (1 - torch.div(1, u_norm + self.eps))
+            u_norm = unit_norm(p)
+            correction = 2.0 * self.norm_loss_factor * (1.0 - torch.div(1, u_norm + self.eps))
             p.mul_(1.0 - lr * correction)
 
             for p in group['params']:
@@ -297,7 +289,7 @@ class Ranger21(Optimizer):
                 max_variance_ma = state['max_variance_ma']
 
                 torch.max(max_variance_ma, variance_ma, out=variance_ma)
-                denom = (variance_ma.sqrt() / math.sqrt(bias_correction2)).add_(group['eps'])
+                de_nom = (variance_ma.sqrt() / math.sqrt(bias_correction2)).add_(group['eps'])
 
                 grad = p.grad
                 grad = centralize_gradient(grad, gc_conv_only=False)
@@ -305,15 +297,15 @@ class Ranger21(Optimizer):
 
                 grad_ma.mul_(beta1 ** 2).add_(grad, alpha=1.0 - beta1 ** 2)
 
-                noise_norm: float = math.sqrt((1 + beta2) ** 2 + beta2 ** 2)
+                noise_norm: float = math.sqrt((1.0 + beta2) ** 2 + beta2 ** 2)
 
                 step_size: float = lr / bias_correction1
 
                 if self.use_softplus:
-                    denom = F.softplus(denom, beta=self.beta_softplus)
+                    de_nom = F.softplus(de_nom, beta=self.beta_softplus)
 
-                pnmomentum = grad_ma.mul(1 + 1.0).add(neg_grad_ma, alpha=-1.0).mul(1.0 / noise_norm)
-                p.addcdiv_(pnmomentum, denom, value=-step_size)
+                pn_momentum = grad_ma.mul(1.0 + 1.0).add(neg_grad_ma, alpha=-1.0).mul(1.0 / noise_norm)
+                p.addcdiv_(pn_momentum, de_nom, value=-step_size)
 
         self.lookahead_process_step()
 
@@ -330,8 +322,8 @@ class Ranger21(Optimizer):
 
                     param_state = self.state[p]
 
-                    p.data.mul_(self.lookahead_blending_alpha).add_(
+                    p.mul_(self.lookahead_blending_alpha).add_(
                         param_state['lookahead_params'],
                         alpha=1.0 - self.lookahead_blending_alpha,
                     )
-                    param_state['lookahead_params'].copy_(p.data)
+                    param_state['lookahead_params'].copy_(p)
