@@ -1,4 +1,4 @@
-from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
+from typing import Any, Callable, Dict, Iterable, Optional, Tuple, Union
 
 import torch
 
@@ -8,4 +8,3 @@ BETAS = Tuple[float, float]
 DEFAULTS = Dict[str, Any]
 PARAMETERS = Optional[Union[Iterable[Dict[str, Any]], Iterable[torch.Tensor]]]
 STATE = Dict[str, Any]
-BUFFER = List[List[Optional[torch.Tensor]]]
