@@ -1,9 +1,12 @@
 import math
-from typing import List, Optional, Tuple, Union
+from typing import Callable, List, Optional, Tuple, Union
 
+import numpy as np
 import torch
 from torch import nn
 from torch.distributed import all_reduce
+from torch.nn import functional as F
+from torch.nn.utils import clip_grad_norm_
 
 from pytorch_optimizer.types import PARAMETERS
 
@@ -34,6 +37,39 @@ def normalize_gradient(x: torch.Tensor, use_channels: bool = False, epsilon: flo
     return x
 
 
+def flatten_grad(grads: List[torch.Tensor]) -> torch.Tensor:
+    return torch.cat([g.flatten() for g in grads])
+
+
+def un_flatten_grad(grads: torch.Tensor, shapes: List[int]) -> List[torch.Tensor]:
+    idx: int = 0
+    un_flatten_grads: List[torch.Tensor] = []
+    for shape in shapes:
+        length = np.prod(shape)
+        un_flatten_grads.append(grads[idx : idx + length].view(shape).clone())
+        idx += length
+    return un_flatten_grads
+
+
+def channel_view(x: torch.Tensor) -> torch.Tensor:
+    return x.view(x.size()[0], -1)
+
+
+def layer_view(x: torch.Tensor) -> torch.Tensor:
+    return x.view(1, -1)
+
+
+def cosine_similarity_by_view(
+    x: torch.Tensor,
+    y: torch.Tensor,
+    eps: float,
+    view_func: Callable[[torch.Tensor], torch.Tensor],
+) -> torch.Tensor:
+    x = view_func(x)
+    y = view_func(y)
+    return F.cosine_similarity(x, y, dim=1, eps=eps).abs_()
+
+
 def clip_grad_norm(parameters: PARAMETERS, max_norm: float = 0, sync: bool = False) -> Union[torch.Tensor, float]:
     """Clips grad norms.
     During combination with FSDP, will also ensure that grad norms are aggregated
@@ -53,22 +89,44 @@ def clip_grad_norm(parameters: PARAMETERS, max_norm: float = 0, sync: bool = Fal
 
     # if syncing we need to manually perform the clipping so that we aggregate properly
     if max_norm > 0 and not sync:
-        return torch.nn.utils.clip_grad_norm_(parameters, max_norm)
+        return clip_grad_norm_(parameters, max_norm)
 
-    norm_sq = sum(p.grad.data.norm() ** 2 for p in parameters if p.grad is not None)
+    norm_sq = sum(p.grad.norm() ** 2 for p in parameters if p.grad is not None)
     if sync:
         # also need to get the norms from all the other sharded works in FSDP
         all_reduce(norm_sq)
 
     grad_norm = math.sqrt(norm_sq)
     if max_norm > 0:
-        clip_coef = max_norm / (grad_norm + 1e-6)
+        clip_coefficient = max_norm / (grad_norm + 1e-6)
         for p in parameters:
-            p.grad.detach().mul_(clip_coef)
+            p.grad.detach().mul_(clip_coefficient)
 
     return grad_norm
 
 
+def projection(
+    p,
+    grad,
+    perturb: torch.Tensor,
+    delta: float,
+    wd_ratio: float,
+    eps: float,
+) -> Tuple[torch.Tensor, float]:
+    wd: float = 1.0
+    expand_size: List[int] = [-1] + [1] * (len(p.shape) - 1)
+    for view_func in (channel_view, layer_view):
+        cosine_sim = cosine_similarity_by_view(grad, p, eps, view_func)
+
+        if cosine_sim.max() < delta / math.sqrt(view_func(p).size()[1]):
+            p_n = p / view_func(p).norm(dim=1).view(expand_size).add_(eps)
+            perturb -= p_n * view_func(p_n * perturb).sum(dim=1).view(expand_size)
+            wd = wd_ratio
+            return perturb, wd
+
+    return perturb, wd
+
+
 def unit_norm(x: torch.Tensor, norm: float = 2.0) -> torch.Tensor:
     keep_dim: bool = True
     dim: Optional[Union[int, Tuple[int, ...]]] = None
